@@ -1,12 +1,15 @@
+"""
+Functions for working with shapefiles.
+"""
 import os
 import collections
-import time
-from functools import partial
 import shutil
 import fiona
 from shapely.geometry import shape, mapping
 import numpy as np
 import pandas as pd
+import pyproj
+from gisutils.projection import get_authority_crs, project
 
 
 def df2shp(dataframe, shpname, geo_column='geometry', index=False,
@@ -46,8 +49,9 @@ def df2shp(dataframe, shpname, geo_column='geometry', index=False,
     """
 
     # first check if output path exists
-    if os.path.split(shpname)[0] != '' and not os.path.isdir(os.path.split(shpname)[0]):
-        raise IOError("Output folder doesn't exist")
+    output_folder = os.path.abspath(os.path.split(shpname)[0])
+    if os.path.split(shpname)[0] != '' and not os.path.isdir(output_folder):
+        raise IOError("Output folder doesn't exist:\n{}".format(output_folder))
 
     # check for empty dataframe
     if len(dataframe) == 0:
@@ -141,7 +145,7 @@ def rename_fields_to_10_characters(columns, limit=10):
 
 
 def shp2df(shplist, index=None, index_dtype=None, clipto=[], filter=None,
-           true_values=None, false_values=None, layer=None,
+           true_values=None, false_values=None, layer=None, dest_crs=None,
            skip_empty_geom=True):
     """Read shapefile/DBF, list of shapefiles/DBFs, or File geodatabase (GDB)
      into pandas DataFrame.
@@ -165,6 +169,21 @@ def shp2df(shplist, index=None, index_dtype=None, clipto=[], filter=None,
         same as argument for pandas read_csv
     layer : str
         Layer name to read (if opening FileGDB)
+    dest_crs : obj
+        A Python int, dict, str, or pyproj.crs.CRS instance
+        passed to the pyproj.crs.from_user_input
+        See http://pyproj4.github.io/pyproj/stable/api/crs/crs.html#pyproj.crs.CRS.from_user_input.
+        Can be any of:
+          - PROJ string
+          - Dictionary of PROJ parameters
+          - PROJ keyword arguments for parameters
+          - JSON string with PROJ parameters
+          - CRS WKT string
+          - An authority string [i.e. 'epsg:4326']
+          - An EPSG integer code [i.e. 4326]
+          - A tuple of ("auth_name": "auth_code") [i.e ('epsg', '4326')]
+          - An object with a `to_wkt` method.
+          - A :class:`pyproj.crs.CRS` class
     skip_empty_geom : True/False, default True
         Drops shapefile entries with null geometries.
         DBF files (which specify null geometries in their schema) will still be read.
@@ -186,79 +205,91 @@ def shp2df(shplist, index=None, index_dtype=None, clipto=[], filter=None,
     else:
         clip = False
 
+    # destination crs for geometries read from shapefile(s)
+    if dest_crs is not None:
+        dest_crs = get_authority_crs(dest_crs)
+
     df = pd.DataFrame()
     for shp in shplist:
         print("\nreading {}...".format(shp))
         if not os.path.exists(shp):
             raise IOError("{} doesn't exist".format(shp))
-        shp_obj = fiona.open(shp, 'r', layer=layer)
 
-        if index is not None:
-            # handle capitolization issues with index field name
-            fields = list(shp_obj.schema['properties'].keys())
-            index = [f for f in fields if index.lower() == f.lower()][0]
+        # crs of current shapefile
+        shp_crs = get_shapefile_crs(shp)
+        # set the destination CRS if none was specified
+        # so that heterogenious shapefiles will be output to
+        # the same CRS
+        if dest_crs is None and shp_crs is not None:
+            dest_crs = shp_crs
 
-        attributes = []
-        # for reading in shapefiles
-        meta = shp_obj.meta
-        if meta['schema']['geometry'] != 'None':
-            if filter is not None:
-                print('filtering on bounding box {}, {}, {}, {}...'.format(*filter))
-            if clip:  # limit what is brought in to items in index of clipto
-                for line in shp_obj.filter(bbox=filter):
-                    props = line['properties']
-                    if not props[index] in clipto:
-                        continue
-                    props['geometry'] = line.get('geometry', None)
-                    attributes.append(props)
-            else:
-                for line in shp_obj.filter(bbox=filter):
-                    props = line['properties']
-                    props['geometry'] = line.get('geometry', None)
-                    attributes.append(props)
-            print('--> building dataframe... (may take a while for large shapefiles)')
-            shp_df = pd.DataFrame(attributes)
-            # reorder fields in the DataFrame to match the input shapefile
-            if len(attributes) > 0:
-                shp_df = shp_df[list(attributes[0].keys())]
+        with fiona.open(shp, 'r', layer=layer) as shp_obj:
 
-            # handle null geometries
-            if len(shp_df) == 0:
-                print('Empty dataframe! No features were read.')
+            if index is not None:
+                # handle capitolization issues with index field name
+                fields = list(shp_obj.schema['properties'].keys())
+                index = [f for f in fields if index.lower() == f.lower()][0]
+
+            attributes = []
+            # for reading in shapefiles
+            meta = shp_obj.meta
+            if meta['schema']['geometry'] != 'None':
                 if filter is not None:
-                    print('Check filter {} for consistency \
-with shapefile coordinate system'.format(filter))
-            # shp_df will only have a geometry column if it isn't empty
-            else:
-                geoms = shp_df.geometry.tolist()
-                if geoms.count(None) == 0:
-                    shp_df['geometry'] = [shape(g) for g in geoms]
-                elif skip_empty_geom:
-                    null_geoms = [i for i, g in enumerate(geoms) if g is None]
-                    shp_df.drop(null_geoms, axis=0, inplace=True)
-                    shp_df['geometry'] = [shape(g) for g in shp_df.geometry.tolist()]
+                    print('filtering on bounding box {}, {}, {}, {}...'.format(*filter))
+                if clip:  # limit what is brought in to items in index of clipto
+                    for line in shp_obj.filter(bbox=filter):
+                        props = line['properties']
+                        if not props[index] in clipto:
+                            continue
+                        props['geometry'] = line.get('geometry', None)
+                        attributes.append(props)
                 else:
-                    shp_df['geometry'] = [shape(g) if g is not None else None
-                                          for g in geoms]
+                    for line in shp_obj.filter(bbox=filter):
+                        props = line['properties']
+                        props['geometry'] = line.get('geometry', None)
+                        attributes.append(props)
+                print('--> building dataframe... (may take a while for large shapefiles)')
+                shp_df = pd.DataFrame(attributes)
+                # reorder fields in the DataFrame to match the input shapefile
+                if len(attributes) > 0:
+                    shp_df = shp_df[list(attributes[0].keys())]
 
-        # for reading in DBF files (just like shps, but without geometry)
-        else:
-            if clip:  # limit what is brought in to items in index of clipto
-                for line in shp_obj:
-                    props = line['properties']
-                    if not props[index] in clipto:
-                        continue
-                    attributes.append(props)
+                # handle null geometries
+                if len(shp_df) == 0:
+                    print('Empty dataframe! No features were read.')
+                    if filter is not None:
+                        print('Check filter {} for consistency \
+    with shapefile coordinate system'.format(filter))
+                # shp_df will only have a geometry column if it isn't empty
+                else:
+                    geoms = shp_df.geometry.tolist()
+                    if geoms.count(None) == 0:
+                        shp_df['geometry'] = [shape(g) for g in geoms]
+                    elif skip_empty_geom:
+                        null_geoms = [i for i, g in enumerate(geoms) if g is None]
+                        shp_df.drop(null_geoms, axis=0, inplace=True)
+                        shp_df['geometry'] = [shape(g) for g in shp_df.geometry.tolist()]
+                    else:
+                        shp_df['geometry'] = [shape(g) if g is not None else None
+                                              for g in geoms]
+
+            # for reading in DBF files (just like shps, but without geometry)
             else:
-                for line in shp_obj:
-                    attributes.append(line['properties'])
-            print('--> building dataframe... (may take a while for large shapefiles)')
-            shp_df = pd.DataFrame(attributes)
-            # reorder fields in the DataFrame to match the input shapefile
-            if len(attributes) > 0:
-                shp_df = shp_df[list(attributes[0].keys())]
+                if clip:  # limit what is brought in to items in index of clipto
+                    for line in shp_obj:
+                        props = line['properties']
+                        if not props[index] in clipto:
+                            continue
+                        attributes.append(props)
+                else:
+                    for line in shp_obj:
+                        attributes.append(line['properties'])
+                print('--> building dataframe... (may take a while for large shapefiles)')
+                shp_df = pd.DataFrame(attributes)
+                # reorder fields in the DataFrame to match the input shapefile
+                if len(attributes) > 0:
+                    shp_df = shp_df[list(attributes[0].keys())]
 
-        shp_obj.close()
         if len(shp_df) == 0:
             continue
         # set the dataframe index from the index column
@@ -267,6 +298,10 @@ with shapefile coordinate system'.format(filter))
                 shp_df[index] = shp_df[index].astype(index_dtype)
             shp_df.index = shp_df[index].values
 
+        # reproject geometries to dest_crs if needed
+        if shp_crs is not None and dest_crs is not None and shp_crs != dest_crs:
+            shp_df['geometry'] = project(shp_df['geometry'], shp_crs, dest_crs)
+
         df = df.append(shp_df)
 
         # convert any t/f columns to numpy boolean data
@@ -316,3 +351,25 @@ def shp_properties(df):
     return properties
 
 
+def get_shapefile_crs(shapefile):
+    """Get the coordinate reference system for a shapefile.
+
+    Parameters
+    ----------
+    shapefile : str
+        Path to a shapefile
+
+    Returns
+    -------
+    crs : pyproj.CRS instance
+
+    """
+    basename, ext = os.path.splitext(shapefile)
+    prjfile = basename + '.prj'
+    if os.path.exists(prjfile):
+        with open(prjfile) as src:
+            wkt = src.read()
+            crs = pyproj.crs.CRS.from_wkt(wkt)
+            return get_authority_crs(crs)
+
+
